import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np

from newton_type import newton_type


# for people who cannot see the plots
if mpl.get_backend() == 'agg':
    mpl.use('WebAgg')
print(f"backend: {mpl.get_backend()}")


tol = 1e-6                 # convergence tolerance
maxit = int(1e4)            # maximum iterations

F = lambda w: w**16 - 2     # function
dF = lambda w: 16 * w**15   # derivative

# visualize function
w = np.linspace(-1,1,100)
plt.figure(1)
plt.plot(w, F(w))
plt.xlabel(r'w')
plt.legend(r'F(w)')
plt.title(r'residual function')

## 3a)

# some initial values
W0 = [.8, 1, 2, 5, 10]

plt.figure(2)
# loop through all values in W0
for w0 in W0:
    # newton method (exact derivatives)
    [W, Res] = newton_type(F, dF, w0, tol, maxit)
    plt.semilogy(np.absolute(Res), label=r'$w0$=%f'%w0)
plt.legend()
plt.xlabel(r'iteration $k$')
plt.ylabel(r'$|F(w_k)|$')
plt.title('convergence of newton''s method for a few initializations')

## 3b)
# analytical bound on local convergence, see solution pdf
w_ana = 2**(-1/15) * 2**(1/16)

# check convergence at these initializations
N = 200
W0 = np.linspace(.9, 2, N)

iters = np.empty(N, dtype=int)
res_fin = np.empty(N, dtype=float)
# loop through all values in W0
for i, w0 in enumerate(W0):
    # newton type with fixed jacobian
    [W, Res] = newton_type(F, dF, w0, tol, maxit)
    iters[i] = len(W) - 1            # number of iterations
    res_fin[i] = np.abs(Res[-1])     # final residual


plt.figure(3)
plt.subplot(2,1,1)
idx = np.argwhere(iters < maxit)
plt.semilogy(W0[idx], iters[idx], 'g.')
idx = np.argwhere(iters >= maxit)
plt.semilogy(W0[idx], iters[idx], 'r.')
plt.semilogy(w_ana*np.array([1, 1]), plt.ylim(), 'k')
plt.legend([r'converged', r'not converged (in maxit)', r'analytic local convergence bound'], loc='lower right')
plt.xlabel(r'$w0$')
plt.ylabel(r'number of iterations')
plt.xlim([W0[0], W0[-1]])
plt.title(r'newton type with fixed Jacobian')

plt.subplot(2,1,2)
plt.semilogy(W0, res_fin, 'b.')
plt.xlim([W0[0], W0[-1]])
plt.semilogy(w_ana*np.array([1, 1]), plt.ylim(), 'k')
plt.xlabel(r'$w0$')
plt.ylabel(r'final residual')

## 3c)
w0 = float(100)
# original version
[W1, Res1] = newton_type(F, dF, w0, tol, maxit)

# lifted version
w0 = w0 * np.ones(4)
# for k = 1:3
#     w0(k+1) = w0(k)^2
# end

Flift = lambda w: [
    w[1] - w[0]**2,
    w[2] - w[1]**2,
    w[3] - w[2]**2,
    -2 + w[3]**2
] 
dFlift = lambda w: [
    [-2*w[0], 1, 0, 0],
    [0, -2*w[1], 1, 0],
    [0, 0, -2*w[2], 1],
    [0, 0, 0, 2*w[3]]
]
[W2, Res2] = newton_type(Flift, dFlift, w0, tol, maxit)

wopt = 2**(1/16)    # exact root, for analysis

plt.figure(4)
plt.semilogy(np.abs(W1 - wopt))
plt.semilogy(np.abs(W2[:,0] - wopt))
plt.xlabel(r'iteration k')
plt.ylabel(r'|w_k - w^*|')
plt.legend([r'original', r'lifted'])
plt.title(r'newton''s method -- direct vs lifted')


plt.show()
